@@ -7,10 +7,29 @@ from http.cookies import SimpleCookie
 from zodbpickle import pickle, pickletools
 
 from test.support import (
-    TestFailed, TESTFN, run_with_locale, no_tracing,
+    TestFailed, TESTFN, run_with_locale,
     _2G, _4G, bigmemtest,
     )
 
+try:
+    from test.support import no_tracing
+except ImportError:
+    from functools import wraps
+    def no_tracing(func):
+        if not hasattr(sys, 'gettrace'):
+            return func
+        @wraps(func)
+        def wrapper(*args, **kwargs):
+            original_trace = sys.gettrace()
+            try:
+                sys.settrace(None)
+                return func(*args, **kwargs)
+            finally:
+                sys.settrace(original_trace)
+        return wrapper
+
+_PY33 = sys.version_info[:2] >= (3, 3)
+
 from zodbpickle.pickle import bytes_types
 
 # Tests that try a number of pickle protocols should have a
@@ -36,6 +55,7 @@ def count_opcode(code, pickle):
             n += 1
     return n
 
+
 class UnseekableIO(io.BytesIO):
     def peek(self, *args):
         raise NotImplementedError
@@ -426,7 +446,7 @@ def create_data():
     x.append(5)
     return x
 
-class AbstractPickleTests:
+class AbstractPickleTests(unittest.TestCase):
     # Subclass must define self.dumps, self.loads.
 
     _testdata = create_data()
@@ -636,13 +656,13 @@ class AbstractPickleTests:
     def test_bytes(self):
         for proto in protocols:
             for s in b'', b'xyz', b'xyz'*100:
-                p = self.dumps(s)
+                p = self.dumps(s, proto)
                 self.assertEqual(self.loads(p), s)
             for s in [bytes([i]) for i in range(256)]:
-                p = self.dumps(s)
+                p = self.dumps(s, proto)
                 self.assertEqual(self.loads(p), s)
             for s in [bytes([i, i]) for i in range(256)]:
-                p = self.dumps(s)
+                p = self.dumps(s, proto)
                 self.assertEqual(self.loads(p), s)
 
     def test_ints(self):
@@ -746,12 +766,14 @@ class AbstractPickleTests:
                 u = self.loads(s)
                 self.assertEqual(t, u)
 
+    @unittest.skipUnless(_PY33, "only for Python >= 3.3")
     def test_ellipsis(self):
         for proto in protocols:
             s = self.dumps(..., proto)
             u = self.loads(s)
             self.assertEqual(..., u)
 
+    @unittest.skipUnless(_PY33, "only for Python >= 3.3")
     def test_notimplemented(self):
         for proto in protocols:
             s = self.dumps(NotImplemented, proto)
@@ -1166,6 +1188,7 @@ class AbstractPickleTests:
         empty = self.loads(b'\x80\x03U\x00q\x00.', encoding='koi8-r')
         self.assertEqual(empty, '')
 
+    @unittest.skipUnless(_PY33, "only for Python >= 3.3")
     def test_int_pickling_efficiency(self):
         # Test compacity of int representation (see issue #12744)
         for proto in protocols:
@@ -1204,7 +1227,7 @@ class AbstractPickleTests:
         dumped = b'\x80\x03X\x01\x00\x00\x00ar\xff\xff\xff\xff.'
         self.assertRaises(ValueError, self.loads, dumped)
 
-class AbstractBytestrTests:
+class AbstractBytestrTests(unittest.TestCase):
     def unpickleEqual(self, data, unpickled):
         loaded = self.loads(data, encoding="bytes")
         self.assertEqual(loaded, unpickled)
@@ -1258,7 +1281,7 @@ class AbstractBytestrTests:
                 b'T,\x01\x00\x00xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxq\x00.',
                 b'x'*300)
 
-class AbstractBytesFallbackTests:
+class AbstractBytesFallbackTests(unittest.TestCase):
     def unpickleEqual(self, data, unpickled):
         loaded = self.loads(data, errors="bytes")
         self.assertEqual(loaded, unpickled)
@@ -1271,7 +1294,65 @@ class AbstractBytesFallbackTests:
                 b"(dp0\nS'y'\np1\nS'\\xff'\np2\nsS'x'\np3\nS'ascii'\np4\ns.",
                 {'x': 'ascii', 'y': b'\xff'})
 
-class BigmemPickleTests:
+class AbstractBytesAsStringTests(unittest.TestCase):
+
+    def pickleEqual(self, proto, data, pickled):
+        dumped = self.dumps(data, proto, bytes_as_strings=True)
+        self.assertEqual(dumped, pickled)
+        # to unpickle these you must use either encoding='bytes' or
+        # errors='bytes'
+        unpickled = self.loads(dumped, errors='bytes')
+        self.assertEqual(unpickled, data)
+
+    def test_save_bytes(self):
+        short_bytestr = b'short byte string \xff'
+        p0_encoded = b"S'" + short_bytestr[:-1] + b"\\xff'\n."
+        p1_encoded = b'U' + bytes([len(short_bytestr)]) + short_bytestr + b'.'
+        p2_encoded = b'\x80\x02' + p1_encoded
+        self.pickleEqual(0, short_bytestr, p0_encoded)
+        self.pickleEqual(1, short_bytestr, p1_encoded)
+        self.pickleEqual(2, short_bytestr, p2_encoded)
+
+    def test_save_bytes_roundtrip(self):
+        for proto in protocols:
+            for ch in range(256):
+                data = bytes([ch])
+                pickled = self.dumps(data, proto, bytes_as_strings=True)
+                unpickled = self.loads(pickled, encoding='bytes')
+                self.assertEqual(unpickled, data)
+
+    def test_dump_dumps(self):
+        # see that the bytes_as_string kwarg is accepted everywhere
+        # (the other tests go throught the Pickler API)
+        # XXX: this is not very abstract; the only way to test the Python
+        # implementation of dump and dumps is to remove the C module
+        f = io.BytesIO()
+        pickle.dump(b'\x00', f, protocol=0, bytes_as_strings=True)
+        self.assertEqual(f.getvalue(), b"S'\\x00'\n.")
+        p = pickle.dumps(b'\x00', protocol=0, bytes_as_strings=True)
+        self.assertEqual(p, b"S'\\x00'\n.")
+
+    def test_save_bytes_roundtrip(self):
+        for proto in protocols:
+            for ch in range(256):
+                data = bytes([ch])
+                pickled = self.dumps(data, proto, bytes_as_strings=True)
+                unpickled = self.loads(pickled, encoding='bytes')
+                self.assertEqual(unpickled, data)
+
+    def test_dump_dumps(self):
+        # see that the bytes_as_string kwarg is accepted everywhere
+        # (the other tests go throught the Pickler API)
+        # XXX: this is not very abstract; the only way to test the Python
+        # implementation of dump and dumps is to remove the C module
+        f = io.BytesIO()
+        pickle.dump(b'\x00', f, protocol=0, bytes_as_strings=True)
+        self.assertEqual(f.getvalue(), b"S'\\x00'\n.")
+        p = pickle.dumps(b'\x00', protocol=0, bytes_as_strings=True)
+        self.assertEqual(p, b"S'\\x00'\n.")
+
+
+class BigmemPickleTests(unittest.TestCase):
 
     # Binary protocols can serialize longs of up to 2GB-1
 
@@ -1435,7 +1516,7 @@ class BadGetattr:
         self.foo
 
 
-class AbstractPickleModuleTests:
+class AbstractPickleModuleTests(unittest.TestCase):
 
     def test_dump_closed_file(self):
         import os
@@ -1495,7 +1576,7 @@ class AbstractPickleModuleTests:
         self.assertRaises(EOFError, pickle.loads, s)
 
 
-class AbstractPersistentPicklerTests:
+class AbstractPersistentPicklerTests(unittest.TestCase):
 
     # This class defines persistent_id() and persistent_load()
     # functions that should be used by the pickler.  All even integers
@@ -1531,7 +1612,7 @@ class AbstractPersistentPicklerTests:
         self.assertEqual(self.load_count, 5)
 
 
-class AbstractPicklerUnpicklerObjectTests:
+class AbstractPicklerUnpicklerObjectTests(unittest.TestCase):
 
     pickler_class = None
     unpickler_class = None
@@ -1682,6 +1763,55 @@ class AbstractPicklerUnpicklerObjectTests:
                 unpickler = self.unpickler_class(f)
                 self.assertEqual(unpickler.load(), data)
 
+    def test_noload_object(self):
+        global _NOLOAD_OBJECT
+        after = {}
+        _NOLOAD_OBJECT = object()
+        aaa = AAA()
+        bbb = BBB()
+        ccc = 1
+        ddd = 1.0
+        eee = ('eee', 1)
+        fff = ['fff']
+        ggg = {'ggg': 0}
+        unpickler = self.unpickler_class
+        f = io.BytesIO()
+        pickler = self.pickler_class(f, protocol=2)
+        pickler.dump(_NOLOAD_OBJECT)
+        after['_NOLOAD_OBJECT'] = f.tell()
+        pickler.dump(aaa)
+        after['aaa'] = f.tell()
+        pickler.dump(bbb)
+        after['bbb'] = f.tell()
+        pickler.dump(ccc)
+        after['ccc'] = f.tell()
+        pickler.dump(ddd)
+        after['ddd'] = f.tell()
+        pickler.dump(eee)
+        after['eee'] = f.tell()
+        pickler.dump(fff)
+        after['fff'] = f.tell()
+        pickler.dump(ggg)
+        after['ggg'] = f.tell()
+        f.seek(0)
+        unpickler = self.unpickler_class(f)
+        unpickler.noload() # read past _NOLOAD_OBJECT
+        self.assertEqual(f.tell(), after['_NOLOAD_OBJECT'])
+        unpickler.noload() # read past aaa
+        self.assertEqual(f.tell(), after['aaa'])
+        unpickler.noload() # read past bbb
+        self.assertEqual(f.tell(), after['bbb'])
+        unpickler.noload() # read past ccc
+        self.assertEqual(f.tell(), after['ccc'])
+        unpickler.noload() # read past ddd
+        self.assertEqual(f.tell(), after['ddd'])
+        unpickler.noload() # read past eee
+        self.assertEqual(f.tell(), after['eee'])
+        unpickler.noload() # read past fff
+        self.assertEqual(f.tell(), after['fff'])
+        unpickler.noload() # read past ggg
+        self.assertEqual(f.tell(), after['ggg'])
+
 
 # Tests for dispatch_table attribute
 
@@ -1694,7 +1824,7 @@ class AAA(object):
 class BBB(object):
     pass
 
-class AbstractDispatchTableTests:
+class AbstractDispatchTableTests(unittest.TestCase):
 
     def test_default_dispatch_table(self):
         # No dispatch_table attribute by default
